# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Filename: decorators.py
# Project: helpers
# Author: Brian Cherinka
# Created: Thursday, 8th October 2020 12:44:20 pm
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2020 Brian Cherinka
# Last Modified: Thursday, 8th October 2020 12:44:21 pm
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import

import inspect
from functools import wraps
from typing import Callable, Type, TypeVar

from sdss_brain import log
from sdss_brain.helpers import get_mapped_version, parse_data_input

T = TypeVar('T')

# global registry of decorators
registry = {}


def register(func):
    ''' Decorator to add class decorators to the global registry '''
    registry[func.__name__] = func
    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)
    return wrapper


def get_parse_input(regex: str = None, keys: list = None, keymap: dict = None,
                    exclude: list = None, include: list = None,
                    order: list = None, delimiter: str = None) -> Callable:
    """ Generate a default parse_input method to be attached to a class """

    def _parse_input(self, value):
        ''' Default parse_input applied with the decorator '''
        akeys = None
        if self.is_access_mixedin and self.path_name is not None:
            akeys = self.access.lookup_keys(self.path_name)
        pkeys = keys or akeys or None

        data = parse_data_input(value, regex=regex, keys=pkeys, keymap=keymap,
                                include=include, exclude=exclude, order=order, delimiter=delimiter)
        log.debug(f'parsing data for {value} returns {data}')

        for k, v in data.items():
            if k != ['filename', 'objectid']:
                setattr(self, k, v)
        return data
    return _parse_input


def create_mapped_properties(kls: Type[T], mapped_version: str):
    ''' Create new read-only properties on a given class

    Creates new read-only properties that extracts a specific version id to an input
    release.  This allows the version id to be updated when the global "release" is
    changed.  See `~sdss_brain.helpers.get_mapped_version` for more details.
    ``mapped_version`` is a mapping key, "[mapping]:[property,]", where [mapping]
    is the name of the key in the "mapped_version" attribute in the brain configuration yaml file,
    and [property,] is a list of version ids to become properties.

    For example a key of "manga:drpver" creates a new read-only property called "drpver" and uses
    `get_mapped_version` to extract the correct version number for a given release from the
    "mapped_version['manga']" key in ~sdss_brain.yaml.

    Parameters
    ----------
        kls : Type
            The class object
        mapped_version : str
            The mapping key to map a specific version onto a release
    '''
    mapkey, attrkey = mapped_version.split(':')
    # set the mapped_version class attribute
    kls.mapped_version = mapkey
    if attrkey:
        # loop over all named values found
        for attr in attrkey.split(','):
            # create read-only property that extracts the correct version number
            # for a given release or returns a valid work version
            setattr(kls, attr, property(lambda self: get_mapped_version(
                kls.mapped_version, release=self.release, key=attr) or self._version.get(attr, None)))


@register
def parser_loader(kls: Type[T] = None, *, pattern: str = None, keys: list = None,
                  keymap: dict = None, exclude: list = None, include: list = None,
                  order: list = None, delimiter: str = None) -> Type[T]:
    """ Class decorator to reduce boilerplate around definition of parse_input method

    Decorator to generate a default `_parse_input` method to be attached to a class.
    The created default method uses `~sdss_brain.helpers.parsing.parse_data_input` to
    match the input string value against the provided regex pattern and returns a dictionary
    containing the parsed "filename" or "objectid". If the pattern is a named-group regex pattern
    the method extracts the names and adds them as attributes on the class.  If no pattern is
    specified, the method checks for available sdss_access template keys and, if found, constructs
    a regex matching pattern from them.  Otherwise, if no keys are found, it performs a greedy
    match and sets the result as the objectid.  The return dictonary adds a `parsed_groups`
    attribute to the instance containing the match group output from the regex match.  This allows
    the user to access the extracted matches when the input pattern is simple, i.e. containing
    no named groups or grouping regex structure.

    Parameters
    ----------
    pattern : str
        The regex pattern to match with
    keys : list
        Optional list of names to construct a named pattern.  Default is to use any sdss_access keys.
    keymap : dict
        Optional dict mapping names to specific patterns to use. Default is None.
    include : list
        A list of access keywords to include in the objectid pattern creation
    exclude: list
        A list of access keywords to exclude in the objectid pattern creation
    order : list
        A list of access keywords to order the objectid pattern by
    delimiter : str
        The delimiter to use when joining keys for the objectid pattern creation

    Returns
    -------
        kls: class
            The decorated class
    """
    def wrap(kls):
        # setup and attach the default parse_input function
        parse_input = get_parse_input(regex=pattern, keys=keys, keymap=keymap, include=include,
                                      exclude=exclude, order=order, delimiter=delimiter)
        setattr(kls, '_parse_input', parse_input)

        # update the __abstractmethod__ with the boilerplate set
        method_set = ['_parse_input']
        kls.__abstractmethods__ = kls.__abstractmethods__.symmetric_difference(method_set)
        return kls

    if not kls:
        return wrap
    return wrap(kls)


def _set_access_path_params(self):
    ''' Default set_access_path_params applied with the decorator'''
    keys = self.access.lookup_keys(self.path_name)
    self.path_params = {k: getattr(self, k) for k in keys}
    log.debug(f'setting new access path_params for {self.path_name}: {self.path_params}')


@register
def access_loader(kls: Type[T] = None, *, name: str = None, defaults: dict = {},
                  mapped_version: str = None) -> Type[T]:
    """ Class decorator to reduce boilerplate around setting of sdss_access parameters

    Decorator to generate a default `_set_access_path_params` method given a template
    path name.  The default method creates an empty `path_params` dictionary using the
    template keywords given a path name.  Default values for template kwargs can be specified
    using the "defaults" argument.

    ``mapped_version`` is a mapping key, "[mapping]:[property,]",
    where [mapping] is the name of the key in the "mapped_version" attribute in the brain
    configuration yaml file, and [property,] is a list of version ids to become properties.  See
    `~sdss_brain.helpers.get_mapped_version` for more details.

    For example a key of "manga:drpver" creates a new read-only property called "drpver" and uses
    `get_mapped_version` to extract the correct version number for a given release from the
    "mapped_version['manga']" key in ~sdss_brain.yaml.

    Parameters
    ----------
        name : str
            The sdss_access template path name
        defaults : dict
            Default values for the sdss_access template keyword arguments
        mapped_version : str
            A mapping key to map a specific version onto a release, e.g. "manga:drpver"

    Returns
    -------
        kls: class
            The decorated class
    """
    def wrap(kls):
        # add the path_name class attribute and add defaults for path_params
        kls.path_name = name
        kls._path_defaults = defaults

        # create new properties for mapped versions
        if mapped_version:
            create_mapped_properties(kls, mapped_version)

        # attach the default set_access_path_params
        setattr(kls, '_set_access_path_params', _set_access_path_params)

        # update the __abstractmethod__ with the boilerplate set
        method_set = ['_set_access_path_params']

        kls.__abstractmethods__ = kls.__abstractmethods__.symmetric_difference(method_set)
        return kls

    if not kls:
        return wrap
    return wrap(kls)


# def sdss_loader(kls=None, *, name=None, defaults={}, mapped_version=None, pattern=None,
#                 include=None, exclude=None, order=None, delimiter=None):
#     """ """

#     def wrap(kls):
#         if name:
#             kls = access_loader(kls=kls, name=name, defaults=defaults,
#                                 mapped_version=mapped_version)

#         parser = any([pattern, include, exclude, order])
#         if parser:
#             kls = parser_loader(kls=kls, pattern=pattern, include=include, exclude=exclude,
#                                 order=order, delimiter=delimiter)
#         return kls

#     if not kls:
#         return wrap

#     return wrap(kls)


def use_decorators(*args):
    ''' Decorator that adds the specified decorators to the function as an attribute '''
    def actual_decorator(func):

        # create a list of the used decorators
        func.registry = list(args)

        @wraps(func)
        def wrapper(*args, **kwargs):
            # append the function to itself as a keyword argument
            return func(*args, **kwargs, self=func)
        return wrapper
    return actual_decorator


@use_decorators('access_loader', 'parser_loader')
def sdss_loader(kls: Type[T] = None, *args: str, **kwargs: str) -> Type[T]:
    """ Class decorator that combines and applies other class decorators

    This decorator applies the decorators specified in `use_decorators` to
    the input class instance.  It effectively combines individual decorators allowing
    a global entry point for the keyword arguments.  This is equivalvent to stacking
    decorators, e.g.::

        @access_loader()
        @parser_loader()
        class B(object):
            pass

    Parameters
    ----------
        args : list
            Any appropriate decorator arguments
        kwargs : list
            Any appropriate decoratoy keyword arguments

    Returns
    -------
        The class that has been decorated

    """

    self = kwargs.get('self', None)
    loaders = self.registry

    def wrap(kls):
        for decorator in loaders:
            # identify the decorator function and get its keyword arguments
            func = registry.get(decorator, None)
            func_kwargs = inspect.getfullargspec(
                func).kwonlyargs or inspect.getfullargspec(func.__wrapped__).kwonlyargs

            # extract any matching input keyword arguments
            found_kwargs = set(func_kwargs) & set(kwargs.keys())
            relevant_kwargs = {k: kwargs[k] for k in found_kwargs}

            # apply the class decorator with relevant kwargs
            kls = func(kls=kls, **relevant_kwargs)

        return kls

    if not kls:
        return wrap

    return wrap(kls)



# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Filename: core.py
# Project: sdss_brain
# Author: Brian Cherinka
# Created: Sunday, 15th March 2020 4:53:35 pm
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2020 Brian Cherinka
# Last Modified: Wednesday, 18th March 2020 4:08:26 pm
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import
import abc
from sdss_brain.config import config
from sdss_brain.exceptions import BrainError
from sdss_brain.mixins.mma import MMAccess, MMAMixIn
from astropy.io import fits


class Base(abc.ABC):
    ''' abstract base class for tools '''

    def __new__(cls, *args, **kwargs):
        # set the correct MMA class
        if MMAccess in cls.mro():
            cls._mma = MMAccess
        else:
            cls._mma = MMAMixIn

        return super().__new__(cls)

    @abc.abstractmethod
    def _load_object_from_file(self, data: object = None) -> None:
        pass

    @abc.abstractmethod
    def _load_object_from_db(self, data=None) -> None:
        pass

    @abc.abstractmethod
    def _load_object_from_api(self, data=None) -> None:
        pass


class HindBrain(Base):
    ''' Base class for utilizing the MMA mixin

    This is a convenience class with the `~sdss_brain.mixins.mma.MMAccess` already implemented.
    This class initializes the ``MMAccess`` and provides logic to load data based
    on the data_origin.  It also provides a simple ``repr``.

    In addition to any abstractmethod from the MMA, this class contains three abstractmethods you
    must override when subclassing.

    - **_load_object_from_file**: defines data load/handling from a local file
    - **_load_object_from_db**: defines data load/handling from a local database
    - **_load_object_from_api**: defines data load/handling from a remote API

    Parameters
    ----------
        data_input : str
            The file or name of target data to load
        filename : str
            The absolute filepath to data to load
        objectid : str
            The object identifier of the data to load
        mode : str
            The operating mode: auto, local, or remote
        release : str
            The data release of the object, e.g. "DR16"
        data : object
            Optional data to instantiate the object with
        download : bool
            If True, downloads the object locally with sdss_access
        ignore_db : bool
            If True, ignores any database connection for local access
        use_db : `~sdssdb.connection.DatabaseConnection`
            a database connection to override the default with

    Attributes
    ----------
        _db : `~sdssdb.connection.DatabaseConnection`
            A relevant sdssdb database connection for the object
        mapped_version : str
            The name of survey/category in the mapped_versions dictionary
    '''
    _db = None
    mapped_version = None
    data_origin = None

    def __new__(cls, *args, **kwargs):
        # set any work versions
        cls.set_work_version(config.work_versions)
        return super().__new__(cls)

    def __init__(self, data_input: str = None, filename: str = None,
                 objectid: str = None, mode: str = None, data: object = None,
                 release: str = None, download: bool = None,
                 ignore_db: bool = None, use_db: bool = None, version: str = None) -> None:

        # set a version for sdsswork data
        checked_release = release or config.release
        if version:
            self.set_work_version(version)
            if checked_release.lower() != 'work':
                raise BrainError('version is only used for "work" data. '
                                 'Please set the input or config release to "WORK"')
        else:
            if not self._version and checked_release.lower() == 'work':
                raise BrainError('You are using a "work" release but have no work versions set! '
                                 'Try setting a global "work_version" dict or specify a "version" input!')

        # initialize the MMA
        self._mma.__init__(self, data_input=data_input, filename=filename,
                           objectid=objectid, mode=mode,
                           release=release, download=download,
                           ignore_db=ignore_db, use_db=use_db or self._db)

        self.data = data
        if self.data_origin == 'file':
            self._load_object_from_file(data=data)
        elif self.data_origin == 'db':
            self._load_object_from_db(data=data)
        elif self.data_origin == 'api':
            self._load_object_from_api()

    def __repr__(self):
        objname = f"objectid='{self.objectid}'" if self.objectid else f"filename='{self.filename}'"
        return (f"<{self.__class__.__name__} {objname}, mode='{self.mode}', "
                f"data_origin='{self.data_origin}'>")

    def __del__(self):
        ''' Destructor for closing open objects '''
        self._close()

    def _close(self):
        ''' close open object for each data_origin '''
        # close open FITS files
        if self.data_origin == 'file' and isinstance(self.data, fits.HDUList):
            self.data.close()
        elif self.data_origin == 'db':
            pass
        elif self.data_origin == 'api':
            pass

    def __enter__(self):
        ''' constructor for context manager '''
        return self

    def __exit__(self, type, value, traceback):
        ''' destructor for context manager '''
        self._close()
        return True

    @classmethod
    def set_work_version(cls, value: dict):
        ''' Set the work version for the given class '''
        if type(value) != dict:
            raise ValueError(f'input verion must be a dictionary')

        # update any existing work versions with the new values
        wv = config.work_versions.copy()
        wv.update(value)
        cls._version = wv


class Brain(HindBrain, MMAccess):
    """ The hind Brain with support for ``sdss_access``

    See `~HindBrain`, `~sdss_brain.mixins.mma.MMAccess`, and `~sdss_brain.mixins.access.AccessMixIn`
    for detailed information.

    """


class BrainNoAccess(HindBrain, MMAMixIn):
    """ A version of `~Brain` without support for ``sdss_access``

    See `~HindBrain` and `~sdss_brain.mixins.mma.MMAMixIn` for detailed
    information.

    """



# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Filename: access.py
# Project: sdss_brain
# Author: Brian Cherinka
# Created: Friday, 2nd October 2020 3:24:02 pm
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2020 Brian Cherinka
# Last Modified: Friday, 2nd October 2020 3:24:02 pm
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import
import abc
import time
import warnings
from functools import wraps
from typing import Type

from sdss_brain import log
from sdss_brain.exceptions import BrainError, BrainMissingDependency, BrainUserWarning
from sdss_brain.config import config

try:
    from sdss_access import Access
except ImportError:
    Access = None

__all__ = ['AccessMixIn']


def create_new_access(release: str) -> Type[Access]:
    ''' create a new sdss_access instance

    Parameters
    ----------
        release : str
            The sdss data release
    '''
    # check for public release
    is_public = 'DR' in release
    rsync_release = release.lower() if is_public else None

    if not Access:
        raise BrainMissingDependency('sdss_access is not installed')

    return Access(public=is_public, release=rsync_release)


def set_access(func):
    ''' Decorator that sets the _access attribute

    Creates a new sdss_access instance if either the _access
    attribute is None or the object release differs from the access
    release.  Ensures that a new sdss_access.Access is instantiated
    when we change releases, e.g. between public DRs or work releases.

    '''
    @wraps(func)
    def wrapper(*args, **kwargs):
        inst = args[0]
        isset = inst._access is not None
        # see if the instance release is different than the access release
        if 'work' in inst.release.lower():
            diffrelease = 'sdsswork' != inst._access.release if isset else None
        else:
            diffrelease = inst.release.lower() != inst._access.release if isset else None
        if not isset or diffrelease:
            inst._access = create_new_access(inst.release)
        return func(*args, **kwargs)
    return wrapper


def check_access_params(func):
    '''Decorator that checks for correct output from set_access_path_params '''

    @wraps(func)
    def wrapper(*args, **kwargs):
        inst = args[0]

        if kwargs.get('force_file', None):
            return func(*args, **kwargs)

        assert hasattr(
            inst, 'path_name'), f'{inst.__class__.__name__} must have a "path_name" class attribute'
        assert hasattr(
            inst, 'path_params'), 'set_access_path_params must set a "path_params" attribute'
        assert getattr(inst, 'path_name'), 'the path_name attribute cannot be None'
        assert getattr(inst, 'path_params'), 'the path_params attribute cannot be None'
        assert type(inst.path_params) == dict, 'the path_params attribute must be a dictionary'
        if inst.filename is None and not all(inst.path_params.values()):
            warnings.warn('Not all path_params are set.  Check how path_params are set '
                          'or for a mismatch between path_params and any extracted parameters '
                          'from _parse_input.  Ensuring any None path_params are set as strings')
            inst.path_params = dict(zip(inst.path_params.keys(), map(str, inst.path_params.values())))
        return func(*args, **kwargs)
    return wrapper


class AccessMixIn(abc.ABC):
    ''' Mixin for implementing multi-modal data access

    This is a class that adds support for dynamic path operations using
    `sdss_access`.  Given a template path name and a defined set of
    template keyword argument, provides convenience methods for constructing
    the full local or url-based pathname, downloading the file with `sdss_access`.
    Also provides the complete `~sdss_access.sync.access.Access` object as a property
    for full range of functionality.  The ``access`` property automatically reconfigures
    itself according to the specified data release on each call.

    This mixin contains one abstractmethod you must override when subclassing.

    - **_set_access_path_params**: sets the arguments needed by `sdss_access`

    Parameters
    ----------
        release : str
            The data release of the object, e.g. "DR16"

    Attributes
    ----------
        path_name : str
            The `sdss_access` template path name
        path_params : dict
            The set of `sdss_access` template path keyword arguments
        access : ~sdss_access.sync.access.Access
            An instance of `sdss_access` using for all path creation and file downloads

    '''

    path_name: str = None

    def __init__(self, *args: str, **kwargs: str):
        self._release = kwargs.get('release', None) or config.release

        # sdss_access attributes
        self._access = None
        self.path_params = None

        self._setup_access()

        super().__init__(*args, **kwargs)

    @property
    @set_access
    def access(self) -> Type[Access]:
        ''' Returns an instance of `~sdss_access.sync.access.Access` '''
        return self._access

    @abc.abstractmethod
    def _set_access_path_params(self) -> None:
        ''' Return the sdss_access path parameters

        This method must be overridden by each subclass and must set at least one
        parameter, "path_params", which specify parameters to be passed
        to sdss_access.  "path_name" must also be specified as a class attribute.

        Attributes
        ----------
            path_name : str
                Required. The sdss_access template path key name.  Must be set on the class.
            path_params : dict
                Required. The keywords needed to fill out the sdss_access template path
        '''

    @check_access_params
    def get_full_path(self, url: str = None, force_file: bool = None) -> str:
        """ Returns the full path of the file in the tree.

        Parameters
        ----------
            url : bool
                If True, specifies the url location rather than the local file location
            force_file : bool
                If True, explicitly returns any set filename attribute instead of constructing
                a path from keyword arguments.

        Returns
        -------
            fullpath : str
                The full path as built by sdss_access
        """

        if force_file:
            return self.filename

        log.debug(f'getting full path for {self.path_name} and params {self.path_params}')
        msg = 'sdss_access was not able to retrieve the full path of the file.'
        fullpath = None
        try:
            if url:
                fullpath = self.access.url(self.path_name, **self.path_params)
            else:
                fullpath = self.access.full(self.path_name, **self.path_params)
        except TypeError as ee:
            warnings.warn(msg + 'Error: {0}'.format(str(ee)), BrainUserWarning)
            raise BrainError(f'Bad input type for sdss_access: {ee}') from ee
        except Exception as ee:
            warnings.warn(msg + 'Error: {0}'.format(str(ee)), BrainUserWarning)
        return fullpath

    def _setup_access(self, params: dict = None, origin: str = None) -> None:
        ''' Set up the initial access parameters

        Sets up an initial default path_params dictionary.  Given a provided `path_name`
        class attribute, looks up the path keyword arguments and creates instance properties,
        as well as a default `path_params` dictionary.  If "params" is specified then properties
        and `path_params` is updated from that input.

        Parameters
        ----------
            params : dict
                A dictionary of access path params
            origin : str
                Indicates the origin of the content calling this method. Either "file" or "object"
        '''

        assert origin in [None, 'file', 'object'], 'origin can only be file or object'

        # do nothing if no path_name set
        if not hasattr(self, 'path_name') or not self.path_name:
            return

        # look up the access keys and create attributes
        keys = self.access.lookup_keys(self.path_name)
        log.debug(f"setting up initial access keys for {keys} for {self.path_name}")
        for k in keys:
            # look up a possible work version
            work_ver = self._version.get(k, None)
            vmsg = ('Version extracted from file is different than your preset "work" '
                    f'version for {k}. Consider updating the configured work version or '
                    'specifying an input version.')

            # look for a default value
            default = self._path_defaults.get(k, None) if hasattr(
                self, '_path_defaults') else None

            # set the work version as default if no default found
            if not default and work_ver and origin == 'object':
                default = work_ver

            # get the attribute value to set
            if params:
                if type(params) != dict:
                    raise TypeError('the path_params attribute must be a dictionary')

                # check if work_ver should supersede the path_param
                if origin == 'object' and work_ver:
                    params[k] = work_ver

                attr_value = params.get(k, default)
            else:
                attr_value = default

            # skip if a class attribute already exists
            if hasattr(self.__class__, k):
                # but first check the work version and issue a warning if there mismatch is found
                if work_ver and origin == 'file' and work_ver != attr_value:
                    warnings.warn(vmsg)
                continue

            # set attributes on the instance
            setattr(self, k, attr_value)

            # issue a warning if the preset work version mismatches from the one
            # extracted from the file
            if origin == 'file' and work_ver and work_ver != getattr(self, k, None):
                warnings.warn(vmsg)

        # create a default path params dictionary
        self.path_params = {k: getattr(self, k) for k in keys}

    @check_access_params
    def download(self) -> None:
        """ Download the file using sdss_access """

        self.access.remote()
        self.access.add(self.path_name, **self.path_params)
        self.access.set_stream()
        self.access.commit()
        paths = self.access.get_paths()
        # adding a millisecond pause for download to finish and file existence to register
        time.sleep(0.001)

        self.filename = paths[0]  # doing this for single files, may need to change



# encoding: utf-8

from sdsstools import get_config, get_logger, get_package_version
from tree import Tree


NAME = 'sdss_brain'

# Loads config
cfg_params = get_config(NAME)

# Inits the logging system. Only shell logging, and exception and warning catching.
# File logging can be started by calling log.start_file_logger(path).
log = get_logger(NAME)
log.setLevel("INFO")


__version__ = get_package_version(path=__file__, package_name=NAME)

tree = Tree()


# !usr/bin/env python
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2017-12-05 12:01:21
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2017-12-05 12:19:32

from __future__ import print_function, division, absolute_import


class BrainError(Exception):
    """A custom core Brain exception"""

    def __init__(self, message=None):

        message = 'There has been an error' \
            if not message else message

        super(BrainError, self).__init__(message)


class BrainNotImplemented(BrainError):
    """A custom exception for not yet implemented features."""

    def __init__(self, message=None):

        message = 'This feature is not implemented yet.' \
            if not message else message

        super(BrainNotImplemented, self).__init__(message)


class BrainAPIError(BrainError):
    """A custom exception for API errors"""

    def __init__(self, message=None):
        if not message:
            message = 'Error with Http Response from Brain API'
        else:
            message = 'Http response error from Brain API. {0}'.format(message)

        super(BrainAPIError, self).__init__(message)


class BrainApiAuthError(BrainAPIError):
    """A custom exception for API authentication errors"""
    pass


class BrainMissingDependency(BrainError):
    """A custom exception for missing dependencies."""
    pass


class BrainWarning(Warning):
    """Base warning for Brain."""


class BrainUserWarning(UserWarning, BrainWarning):
    """The primary warning class."""
    pass


class BrainSkippedTestWarning(BrainUserWarning):
    """A warning for when a test is skipped."""
    pass


class BrainDeprecationWarning(BrainUserWarning):
    """A warning for deprecated features."""
    pass



# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Filename: config.py
# Project: sdss_brain
# Author: Brian Cherinka
# Created: Friday, 14th February 2020 1:41:34 pm
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2020 Brian Cherinka
# Last Modified: Tuesday, 17th March 2020 4:20:37 pm
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import
from sdss_brain import cfg_params, log, tree
from sdss_brain.exceptions import BrainError


class Config(object):
    ''' Main configuration class for SDSS '''

    def __init__(self):
        self._mode = 'auto'
        self._release = None
        self.download = False
        self.ignore_db = False
        self.work_versions = None

        # load default config parameters
        self._load_defaults()

        # get allowed releases from the Tree
        self._allowed_releases = tree.get_available_releases()

        # set a default release or get the latest one
        default_release = self._custom_config.get('default_release', None)
        self.release = default_release or self._get_latest_release()

        # set any work versions from the config
        self.set_work_versions()

    def __repr__(self):
        return f'<SDSSConfig(release={self.release}, mode={self.mode})>'

    @property
    def mode(self) -> str:
        return self._mode

    @mode.setter
    def mode(self, value: str):
        assert value in ['local', 'remote', 'auto'], ('config.mode must be "local",'
                                                      '"remote", or "auto".')
        self._mode = value

    @property
    def release(self) -> str:
        return self._release

    @release.setter
    def release(self, value: str) -> None:
        value = value.upper()
        if value not in self._allowed_releases:
            raise BrainError(f'trying to set an invalid release version {value}. '
                             f'Valid releases are: {", ".join(self._allowed_releases)}')

        # replant the tree
        if value.lower() == 'work':
            tree.replant_tree('sdsswork')
        else:
            tree.replant_tree(value.lower())

        self._release = value

    def set_release(self, version: str = None) -> None:
        ''' Set a new release

        If version not specified, uses the latest public DR

        Parameters:
            version (str):
                The new data release to set
        '''
        if not version:
            version = self._get_latest_release()
            log.info(f'Setting release to latest: {version}')
        self.release = version

    def _get_latest_release(self) -> None:
        ''' get the latest public DR release '''
        drsonly = [i for i in self._allowed_releases if 'DR' in i]
        return max(drsonly, key=lambda t: int(t.rsplit('DR', 1)[-1]))

    def list_allowed_releases(self) -> None:
        ''' list the allowed releases based on the available tree environment configurations '''
        return self._allowed_releases

    def _load_defaults(self) -> None:
        ''' Load the Brain config yaml file and update any parameters '''

        # update any matching Config values
        for key, value in cfg_params.items():
            if hasattr(self, key):
                self.__setattr__(key, value)

        self._custom_config = cfg_params

    def set_work_versions(self, values: dict = {}):
        """ Set the versions used for sdsswork

        Sets the versions used by sdswork globally into the config.
        Input is a valid dictionary containing version names and numbers as
        key value pairs, e.g. `{'drpver':'v2_4_3', 'run2d':'v1_1_1', 'apred':'r12'}`.
        Optionally can set them permanently in the custom configuration YAML file.

        The input dictionary is merged with any values specified from the custom config.

        Parameters
        ----------
        values : dict, optional
            Dictionary of version names:numbers needed in paths, by default {}

        Raises
        ------
        ValueError
            when the input is not a dictionary
        """

        if type(values) != dict:
            raise ValueError('input versions must be a dict')

        cfg_work = self._custom_config.get('work_versions', {})
        cfg_work.update(values)
        self.work_versions = cfg_work


config = Config()




# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Filename: cubes.py
# Project: tools
# Author: Brian Cherinka
# Created: Sunday, 11th October 2020 3:15:17 pm
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2020 Brian Cherinka
# Last Modified: Sunday, 11th October 2020 3:15:17 pm
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import

from sdss_brain.tools import Spectrum
from sdss_brain.helpers import sdss_loader


class Cube(Spectrum):
    """ Base class for cubes """
    pass


@sdss_loader(name='mangacube', defaults={'wave': 'LOG'}, mapped_version='manga:drpver',
             pattern=r'(?P<plateifu>(?P<plate>\d{4,5})-(?P<ifu>\d{3,5}))|(?P<mangaid>\d{1,2}-\d{4,9})')
class MangaCube(Spectrum):
    """ Class representing a MaNGA IFU datacube for a single galaxy """
    specutils_format: str = 'MaNGA cube'



# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Filename: spectra.py
# Project: tools
# Author: Brian Cherinka
# Created: Sunday, 11th October 2020 3:15:12 pm
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2020 Brian Cherinka
# Last Modified: Sunday, 11th October 2020 3:15:12 pm
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import

import warnings

try:
    import matplotlib.pyplot as plt
except ImportError:
    plt = None
from astropy.io.registry import IORegistryError
from specutils import Spectrum1D
from typing import Type

from sdss_brain.core import Brain
from sdss_brain.exceptions import BrainNotImplemented, BrainMissingDependency
from sdss_brain.helpers import (sdss_loader, get_mapped_version, load_fits_file, parse_data_input,
                                load_from_url)


class Spectrum(Brain):
    ''' Base class for working with spectral data using specutils

    Uses `specutils <https://specutils.readthedocs.io/en/stable/>`_ to load the
    underlying spectral data into an Astropy Spectrum1D object, which is represented as an
    nd-data array object.

    Attributes
    ----------
        spectrum : `~specutils.Spectrum1D`
            An Astropy specutils Spectrum1D object
    '''
    spectrum: Type[Spectrum1D] = None
    specutils_format: str = None

    def _load_object_from_file(self, data=None) -> None:
        self.data = data or load_fits_file(self.filename)
        self.header = self.data['PRIMARY'].header
        self._load_spectrum()

    def _load_object_from_db(self, data=None) -> None:
        raise BrainNotImplemented('This method must be implemented by the user')

    def _load_object_from_api(self, data=None) -> None:
        # for now, do a simple get request to grab the file into memory
        self.data = data or load_from_url(self.get_full_path(url=True))
        self.header = self.data['PRIMARY'].header
        self._load_spectrum()

    def _load_spectrum(self) -> None:
        ''' Load the `~specutils.Spectrum1D` object '''
        try:
            # check robustness of this to self.data/self.filename when specutils 1.1.1 is released
            self.spectrum = Spectrum1D.read(self.data, format=self.specutils_format)
        except IORegistryError:
            warnings.warn('Could not load Spectrum1D for format '
                          f'{self.specutils_format}, {self.filename}')

    def plot(self, *args, ax=None, x_label: str = 'Wavelength', y_label: str = 'Flux',
             title: str = None, **kwargs):
        """ A simple quick matplotlib plot of the spectrum

        Create a quickplot matplotlib line profile plot for spectra

        Parameters
        ----------
        args : int|list, optional
            for multi-dimensional spectra, the index to plot, by default None
        ax : object, optional
            An existing matplotlib Axes object by default None
        x_label : str, optional
            The x-axis plot label, by default 'Wavelength'
        y_label : str, optional
            The y-axis plot label, by default 'Flux'
        title : str, optional
            The plot title, by default None

        Returns
        -------
        matplotlib.plt.Axes
            The matplotlib Axes object

        Raises
        ------
        BrainMissingDependency
            when matplotlib is not installed
        """
        if not self.spectrum:
            return

        if not plt:
            raise BrainMissingDependency("Package matplotlib not installed.")

        if self.spectrum.flux.ndim > 1:
            index = args
            if not index:
                raise ValueError('The spectrum is n-dimensional.  You must specify a spectrum '
                                 'index to plot.')

            if not all([type(i) == int for i in index]):
                raise ValueError('Input spectrum indices must be integers')

            min_index = self.spectrum.flux.ndim - 1
            index = index if isinstance(index, (tuple, list)) else [index] if type(index) == int else index
            if len(index) < min_index:
                raise ValueError(f'The spectrum has dimension={self.spectrum.flux.ndim}. '
                                 f'You must specify a minimum of {min_index} indices')

            if min_index == 1:
                ii = index[0]
                ydata = self.spectrum.flux[ii]
            elif min_index == 2:
                ii, jj = index
                ydata = self.spectrum.flux[ii, jj]
            else:
                raise ValueError('plot cannot currently support spectral dimensions higher than 3')
        else:
            ydata = self.spectrum.flux

        if not ax:
            ax = plt.gca()

        title = title or f'Object: {self.objectid or self.filename.stem}'
        ax.plot(self.spectrum.wavelength, ydata, **kwargs)
        ax.set_ylabel(f'{y_label} [{self.spectrum.flux.unit.to_string(format="latex_inline")}]')
        ax.set_xlabel(f'{x_label} [{self.spectrum.wavelength.unit.to_string(format="latex_inline")}]')
        ax.set_title(title)
        return ax


# example of using a custom pattern in the sdss_loader
@sdss_loader(name='spec-lite', mapped_version='eboss:run2d',
             pattern=r'(?P<plateid>\d{4,5})-(?P<mjd>\d{5})-(?P<fiberid>\d{1,4})')
class Eboss(Spectrum):
    """ Class representing a single fiber SDSS spectra from BOSS/EBOSS

    Parameters
    ----------
        lite : bool
            If True, loads the "spec-lite" spectral data.  When False, loads
            the "full" spectral data.Default is True.
    """
    specutils_format: str = 'SDSS-III/IV spec'

    def __init__(self, *args: str, lite: bool = True, **kwargs: str) -> None:
        self.lite = lite
        self.path_name = f'spec{"-lite" if lite else ""}'
        super(Eboss, self).__init__(*args, **kwargs)

    def __repr__(self):
        old = super(Eboss, self).__repr__()
        return old.replace('>', f', lite={self.lite}>')


# example of using a access keys to define the pattern in the sdss_loader
# setting a custom delimiter to "--" since APOGEE field names can have "-" in them.
@sdss_loader(name='apStar', defaults={'apstar': 'stars', 'prefix': 'ap'}, delimiter='--',
             mapped_version='apogee:apred', order=['telescope', 'field', 'obj'])
class ApStar(Spectrum):
    """ Class representing an APOGEE combined spectrum for a single star """
    specutils_format: str = 'APOGEE apStar'


@sdss_loader(name='apVisit', defaults={'prefix': 'ap'}, delimiter='--',
             mapped_version='apogee:apred', order=['telescope', 'field', 'plate', 'mjd', 'fiber'])
class ApVisit(Spectrum):
    """ Class representing an APOGEE single visit spectrum for a given star """
    specutils_format: str = 'APOGEE apVisit'


# example of overloading the methods manually
class AspcapStar(Spectrum):
    """ Class representing an APOGEE spectrum for a single star with ASPCAP results """
    specutils_format: str = 'APOGEE aspcapStar'
    path_name: str = 'aspcapStar'
    mapped_version: str = 'apogee'

    def _parse_input(self, value):
        # use the sdss_access keys to form the object id and parse it
        keys = self.access.lookup_keys(self.path_name)
        data = parse_data_input(value, keys=keys, delimiter='--',
                                order=['aspcap', 'telescope', 'field', 'obj'], )
        return data

    def _set_access_path_params(self):
        # extract the apred version id based on the data release
        apred = get_mapped_version(self.mapped_version, release=self.release, key='apred')

        # set the path params using the instance attributes extracted from _parse_input
        self.path_params = {'telescope': self.telescope, 'apred': apred,
                            'field': self.field, 'obj': self.obj, 'aspcap': self.aspcap}
